import uuid

import pytest
from gcloud.aio.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.aio.storage import Storage

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session


@pytest.mark.asyncio
async def test_metadata_multipart(bucket_name, creds):
    object_name = f'{uuid.uuid4().hex}/{uuid.uuid4().hex}.txt'
    original_data = f'{uuid.uuid4().hex}'
    original_metadata = {'Content-Disposition': 'inline',
                         'metadata':
                         {'a': 1,
                          'b': 2,
                          'c': [1, 2, 3],
                          'd': {'a': 4, 'b': 5}}}
    # Google casts all metadata elements as string.
    google_metadata = {'Content-Disposition': 'inline',
                       'metadata':
                       {'a': str(1),
                        'b': str(2),
                        'c': str([1, 2, 3]),
                        'd': str({'a': 4, 'b': 5})}}

    async with Session() as session:
        storage = Storage(service_file=creds, session=session)

        # Without metadata
        res0 = await storage.upload(bucket_name, object_name, original_data,
                                    force_resumable_upload=False)
        data0 = await storage.download(bucket_name, res0['name'])
        await storage.download_metadata(bucket_name, res0['name'])

        # With metadata
        res = await storage.upload(bucket_name, object_name, original_data,
                                   metadata=original_metadata)
        data = await storage.download(bucket_name, res['name'])
        data_metadata = await storage.download_metadata(bucket_name,
                                                        res['name'])

        assert res['name'] == object_name
        assert str(data, 'utf-8') == original_data
        assert data == data0

        assert data_metadata.pop('contentDisposition') == 'inline'
        assert data_metadata['metadata'] == google_metadata['metadata']


@pytest.mark.asyncio
async def test_metadata_resumable(bucket_name, creds):
    object_name = f'{uuid.uuid4().hex}/{uuid.uuid4().hex}.txt'
    original_data = f'{uuid.uuid4().hex}'
    original_metadata = {'Content-Disposition': 'inline',
                         'metadata':
                         {'a': 1,
                          'b': 2,
                          'c': [1, 2, 3],
                          'd': {'a': 4, 'b': 5}}}
    # Google casts all metadata elements as string.
    google_metadata = {'Content-Disposition': 'inline',
                       'metadata':
                       {'a': str(1),
                        'b': str(2),
                        'c': str([1, 2, 3]),
                        'd': str({'a': 4, 'b': 5})}}

    async with Session() as session:
        storage = Storage(service_file=creds, session=session)

        # Without metadata
        res0 = await storage.upload(bucket_name, object_name, original_data,
                                    force_resumable_upload=True)
        data0 = await storage.download(bucket_name, res0['name'])
        await storage.download_metadata(bucket_name, res0['name'])

        # With metadata
        res = await storage.upload(bucket_name, object_name, original_data,
                                   metadata=original_metadata,
                                   force_resumable_upload=True)
        data = await storage.download(bucket_name, res['name'])
        data_metadata = await storage.download_metadata(bucket_name,
                                                        res['name'])

        assert res['name'] == object_name
        assert str(data, 'utf-8') == original_data
        assert data == data0

        assert data_metadata.pop('contentDisposition') == 'inline'
        assert data_metadata['metadata'] == google_metadata['metadata']


@pytest.mark.asyncio
async def test_metadata_copy(bucket_name, creds):
    object_name = f'{uuid.uuid4().hex}/{uuid.uuid4().hex}.txt'
    copied_object_name = f'{object_name}.copy'
    original_data = f'{uuid.uuid4().hex}'
    original_metadata = {'Content-Disposition': 'inline',
                         'metadata':
                         {'a': 1,
                          'b': 2,
                          'c': [1, 2, 3],
                          'd': {'a': 4, 'b': 5}}}
    # Google casts all metadata elements as string.
    google_metadata = {'Content-Disposition': 'inline',
                       'metadata':
                       {'a': str(1),
                        'b': str(2),
                        'c': str([1, 2, 3]),
                        'd': str({'a': 4, 'b': 5})}}

    async with Session() as session:
        storage = Storage(service_file=creds, session=session)

        # Without metadata
        res0 = await storage.upload(bucket_name, object_name, original_data,
                                    force_resumable_upload=True)
        data0 = await storage.download(bucket_name, res0['name'])

        await storage.copy(bucket_name, object_name, bucket_name,
                           new_name=copied_object_name,
                           metadata=original_metadata)

        data = await storage.download(bucket_name, copied_object_name)
        data_metadata = await storage.download_metadata(
            bucket_name, copied_object_name
        )

        assert data == data0

        assert data_metadata.pop('contentDisposition') == 'inline'
        assert data_metadata['metadata'] == google_metadata['metadata']



import json
import uuid

import pytest
from gcloud.aio.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.aio.storage import Storage

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import HTTPError as ResponseError
    from requests import Session
else:
    from aiohttp import ClientError as ResponseError
    from aiohttp import ClientSession as Session


@pytest.mark.asyncio
@pytest.mark.parametrize('uploaded_data,expected_data,file_extension', [
    ('test', b'test', 'txt'),
    (b'test', b'test', 'bin'),
    (json.dumps({'data': 1}), json.dumps({'data': 1}).encode('utf-8'), 'json'),
])
async def test_object_life_cycle(bucket_name, creds, uploaded_data,
                                 expected_data, file_extension):
    object_name = f'{uuid.uuid4().hex}/{uuid.uuid4().hex}.{file_extension}'
    copied_object_name = f'copyof_{object_name}'

    async with Session() as session:
        storage = Storage(service_file=creds, session=session)
        await storage.upload(bucket_name, object_name, uploaded_data)

        bucket = storage.get_bucket(bucket_name)
        blob = await bucket.get_blob(object_name)
        constructed_result = await blob.download()
        assert constructed_result == expected_data

        direct_result = await storage.download(bucket_name, object_name)
        assert direct_result == expected_data

        await storage.copy(bucket_name, object_name, bucket_name,
                           new_name=copied_object_name)

        direct_result = await storage.download(bucket_name, copied_object_name)
        assert direct_result == expected_data

        await storage.delete(bucket_name, object_name)
        await storage.delete(bucket_name, copied_object_name)

        with pytest.raises(ResponseError):
            await storage.download(bucket_name, object_name)

        with pytest.raises(ResponseError):
            await storage.download(bucket_name, copied_object_name)



import binascii
import enum
import io
import json
import logging
import mimetypes
import os
import sys
from typing import Any
from typing import AnyStr
from typing import Dict
from typing import IO
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from urllib.parse import quote

from gcloud.aio.auth import AioSession  # pylint: disable=no-name-in-module
from gcloud.aio.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.aio.auth import Token  # pylint: disable=no-name-in-module
from gcloud.aio.storage.bucket import Bucket
from gcloud.aio.storage.constants import DEFAULT_TIMEOUT

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from time import sleep
    from requests import HTTPError as ResponseError
    from requests import Session
    from builtins import open as file_open
else:
    from aiofiles import open as file_open  # type: ignore[no-redef]
    from asyncio import sleep  # type: ignore[misc]
    from aiohttp import ClientResponseError as ResponseError  # type: ignore[no-redef]  # pylint: disable=line-too-long
    from aiohttp import ClientSession as Session  # type: ignore[no-redef]

API_ROOT = 'https://www.googleapis.com/storage/v1/b'
API_ROOT_UPLOAD = 'https://www.googleapis.com/upload/storage/v1/b'
VERIFY_SSL = True
SCOPES = [
    'https://www.googleapis.com/auth/devstorage.read_write',
]

MAX_CONTENT_LENGTH_SIMPLE_UPLOAD = 5 * 1024 * 1024  # 5 MB

STORAGE_EMULATOR_HOST = os.environ.get('STORAGE_EMULATOR_HOST')
if STORAGE_EMULATOR_HOST:
    API_ROOT = f'https://{STORAGE_EMULATOR_HOST}/storage/v1/b'
    API_ROOT_UPLOAD = f'https://{STORAGE_EMULATOR_HOST}/upload/storage/v1/b'
    VERIFY_SSL = False

log = logging.getLogger(__name__)


def choose_boundary() -> str:
    """Stolen from urllib3.filepost.choose_boundary() as of v1.26.2."""
    boundary = binascii.hexlify(os.urandom(16))
    if sys.version_info.major == 2:
        return boundary  # type: ignore[return-value]
    return boundary.decode('ascii')


def encode_multipart_formdata(fields: List[Tuple[Dict[str, str], bytes]],
                              boundary: str) -> Tuple[bytes, str]:
    """
    Stolen from urllib3.filepost.encode_multipart_formdata() as of v1.26.2.

    Very heavily modified to be compatible with our gcloud-rest converter and
    to avoid unnecessary urllib3 dependencies (since that's only included with
    requests, not aiohttp).
    """
    body: List[bytes] = []
    for headers, data in fields:
        body.append(f'--{boundary}\r\n'.encode('utf-8'))

        # The below is from RequestFields.render_headers()
        # Since we only use Content-Type, we could simplify the below to a
        # single line... but probably best to be safe for future modifications.
        for field in ['Content-Disposition', 'Content-Type',
                      'Content-Location']:
            value = headers.pop(field, None)
            if value:
                body.append(f'{field}: {value}\r\n'.encode('utf-8'))
        for field, value in headers.items():
            # N.B. potential bug copied from urllib3 code; zero values should
            # be sent! Keeping it for now, since Google libs use urllib3 for
            # their examples.
            if value:
                body.append(f'{field}: {value}\r\n'.encode('utf-8'))

        body.append(b'\r\n')
        body.append(data)
        body.append(b'\r\n')

    body.append(f'--{boundary}--\r\n'.encode('utf-8'))

    # N.B. 'multipart/form-data' in upstream, but Google wants 'related'
    content_type = f'multipart/related; boundary={boundary}'

    return b''.join(body), content_type


class UploadType(enum.Enum):
    SIMPLE = 1
    RESUMABLE = 2
    MULTIPART = 3  # unused: SIMPLE upgrades to MULTIPART when metadata exists


class StreamResponse:
    """This class provides an abstraction between the slightly different
    recommended streaming implementations between requests and aiohttp.
    """
    def __init__(self, response: Any) -> None:
        self._response = response
        self._iter: Optional[Iterator[bytes]] = None

    async def read(self, size: int = -1) -> bytes:
        chunk: bytes
        if BUILD_GCLOUD_REST:
            if self._iter is None:
                self._iter = self._response.iter_content(chunk_size=size)
            chunk = next(self._iter, b'')
        else:
            chunk = await self._response.content.read(size)
        return chunk


class Storage:
    def __init__(self, *,
                 service_file: Optional[Union[str, IO[AnyStr]]] = None,
                 token: Optional[Token] = None,
                 session: Optional[Session] = None) -> None:
        self.session = AioSession(session, verify_ssl=VERIFY_SSL)
        self.token = token or Token(service_file=service_file, scopes=SCOPES,
                                    session=self.session.session)

    async def _headers(self) -> Dict[str, str]:
        if STORAGE_EMULATOR_HOST:
            return {}

        token = await self.token.get()
        return {
            'Authorization': f'Bearer {token}',
        }

    def get_bucket(self, bucket_name: str) -> Bucket:
        return Bucket(self, bucket_name)

    # pylint: disable=too-many-locals
    async def copy(self, bucket: str, object_name: str,
                   destination_bucket: str, *, new_name: Optional[str] = None,
                   metadata: Optional[Dict[str, Any]] = None,
                   params: Optional[Dict[str, str]] = None,
                   headers: Optional[Dict[str, str]] = None,
                   timeout: int = DEFAULT_TIMEOUT,
                   session: Optional[Session] = None) -> Dict[str, Any]:

        """
        When files are too large, multiple calls to `rewriteTo` are made. We
        refer to the same copy job by using the `rewriteToken` from the
        previous return payload in subsequent `rewriteTo` calls.

        Using the `rewriteTo` GCS API is preferred in part because it is able
        to make multiple calls to fully copy an object whereas the `copyTo` GCS
        API only calls `rewriteTo` once under the hood, and thus may fail if
        files are large.

        In the rare case you need to resume a copy operation, include the
        `rewriteToken` in the `params` dictionary. Once you begin a multi-part
        copy operation, you then have 1 week to complete the copy job.

        https://cloud.google.com/storage/docs/json_api/v1/objects/rewrite
        """
        if not new_name:
            new_name = object_name

        url = (f"{API_ROOT}/{bucket}/o/{quote(object_name, safe='')}/rewriteTo"
               f"/b/{destination_bucket}/o/{quote(new_name, safe='')}")

        # We may optionally supply metadata* to apply to the rewritten
        # object, which explains why `rewriteTo` is a POST endpoint; when no
        # metadata is given, we have to send an empty body.
        # * https://cloud.google.com/storage/docs/json_api/v1/objects#resource
        metadict = (metadata or {}).copy()
        metadict = {self._format_metadata_key(k): v
                    for k, v in metadict.items()}
        if 'metadata' in metadict:
            metadict['metadata'] = {
                str(k): str(v) if v is not None else None
                for k, v in metadict['metadata'].items()}

        metadata_ = json.dumps(metadict)

        headers = headers or {}
        headers.update(await self._headers())
        headers.update({
            'Content-Length': str(len(metadata_)),
            'Content-Type': 'application/json; charset=UTF-8',
        })

        params = params or {}

        s = AioSession(session) if session else self.session
        resp = await s.post(url, headers=headers, params=params,
                            timeout=timeout, data=metadata_)

        data: Dict[str, Any] = await resp.json(content_type=None)

        while not data.get('done') and data.get('rewriteToken'):
            params['rewriteToken'] = data['rewriteToken']
            resp = await s.post(url, headers=headers, params=params,
                                timeout=timeout)
            data = await resp.json(content_type=None)

        return data

    async def delete(self, bucket: str, object_name: str, *,
                     timeout: int = DEFAULT_TIMEOUT,
                     params: Optional[Dict[str, str]] = None,
                     headers: Optional[Dict[str, str]] = None,
                     session: Optional[Session] = None) -> str:
        # https://cloud.google.com/storage/docs/request-endpoints#encoding
        encoded_object_name = quote(object_name, safe='')
        url = f'{API_ROOT}/{bucket}/o/{encoded_object_name}'
        headers = headers or {}
        headers.update(await self._headers())

        s = AioSession(session) if session else self.session
        resp = await s.delete(url, headers=headers, params=params or {},
                              timeout=timeout)

        try:
            data: str = await resp.text()
        except (AttributeError, TypeError):
            data = str(resp.text)

        return data

    async def download(self, bucket: str, object_name: str, *,
                       headers: Optional[Dict[str, Any]] = None,
                       timeout: int = DEFAULT_TIMEOUT,
                       session: Optional[Session] = None) -> bytes:
        return await self._download(bucket, object_name, headers=headers,
                                    timeout=timeout, params={'alt': 'media'},
                                    session=session)

    async def download_to_filename(self, bucket: str, object_name: str,
                                   filename: str, **kwargs: Any) -> None:
        async with file_open(  # type: ignore[attr-defined]
            filename,
            mode='wb+',
        ) as file_object:
            await file_object.write(
                await self.download(bucket, object_name, **kwargs)
            )

    async def download_metadata(self, bucket: str, object_name: str, *,
                                headers: Optional[Dict[str, Any]] = None,
                                session: Optional[Session] = None,
                                timeout: int = DEFAULT_TIMEOUT
                                ) -> Dict[str, Any]:
        data = await self._download(bucket, object_name, headers=headers,
                                    timeout=timeout, session=session)
        metadata: Dict[str, Any] = json.loads(data.decode())
        return metadata

    async def download_stream(self, bucket: str, object_name: str, *,
                              headers: Optional[Dict[str, Any]] = None,
                              timeout: int = DEFAULT_TIMEOUT,
                              session: Optional[Session] = None
                              ) -> StreamResponse:
        """Download a GCS object in a buffered stream.

        Args:
            bucket (str): The bucket from which to download.
            object_name (str): The object within the bucket to download.
            headers (Optional[Dict[str, Any]], optional): Custom header values
                for the request, such as range. Defaults to None.
            timeout (int, optional): Timeout, in seconds, for the request. Note
                that with this function, this is the time to the beginning of
                the response data (TTFB). Defaults to 10.
            session (Optional[Session], optional): A specific session to
                (re)use. Defaults to None.

        Returns:
            StreamResponse: A object encapsulating the stream, similar to
            io.BufferedIOBase, but it only supports the read() function.
        """
        return await self._download_stream(bucket, object_name,
                                           headers=headers, timeout=timeout,
                                           params={'alt': 'media'},
                                           session=session)

    async def list_objects(self, bucket: str, *,
                           params: Optional[Dict[str, str]] = None,
                           headers: Optional[Dict[str, Any]] = None,
                           session: Optional[Session] = None,
                           timeout: int = DEFAULT_TIMEOUT) -> Dict[str, Any]:
        url = f'{API_ROOT}/{bucket}/o'
        headers = headers or {}
        headers.update(await self._headers())

        s = AioSession(session) if session else self.session
        resp = await s.get(url, headers=headers, params=params or {},
                           timeout=timeout)
        data: Dict[str, Any] = await resp.json(content_type=None)
        return data

    # https://cloud.google.com/storage/docs/json_api/v1/how-tos/upload
    # pylint: disable=too-many-locals
    async def upload(self, bucket: str, object_name: str, file_data: Any,
                     *, content_type: Optional[str] = None,
                     parameters: Optional[Dict[str, str]] = None,
                     headers: Optional[Dict[str, str]] = None,
                     metadata: Optional[Dict[str, Any]] = None,
                     session: Optional[Session] = None,
                     force_resumable_upload: Optional[bool] = None,
                     timeout: int = 30) -> Dict[str, Any]:
        url = f'{API_ROOT_UPLOAD}/{bucket}/o'

        stream = self._preprocess_data(file_data)

        if BUILD_GCLOUD_REST and isinstance(stream, io.StringIO):
            # HACK: `requests` library does not accept `str` as `data` in `put`
            # HTTP request.
            stream = io.BytesIO(stream.getvalue().encode('utf-8'))

        content_length = self._get_stream_len(stream)

        # mime detection method same as in aiohttp 3.4.4
        content_type = content_type or mimetypes.guess_type(object_name)[0]

        parameters = parameters or {}

        headers = headers or {}
        headers.update(await self._headers())
        headers.update({
            'Content-Length': str(content_length),
            'Content-Type': content_type or '',
        })

        upload_type = self._decide_upload_type(force_resumable_upload,
                                               content_length)
        log.debug('using %r gcloud storage upload method', upload_type)

        if upload_type == UploadType.RESUMABLE:
            return await self._upload_resumable(
                url, object_name, stream, parameters, headers,
                metadata=metadata, session=session, timeout=timeout)
        if upload_type == UploadType.SIMPLE:
            if metadata:
                return await self._upload_multipart(
                    url, object_name, stream, parameters, headers, metadata,
                    session=session, timeout=timeout)
            return await self._upload_simple(
                url, object_name, stream, parameters, headers, session=session,
                timeout=timeout)

        raise TypeError(f'upload type {upload_type} not supported')

    async def upload_from_filename(self, bucket: str, object_name: str,
                                   filename: str,
                                   **kwargs: Any) -> Dict[str, Any]:
        async with file_open(  # type: ignore[attr-defined]
            filename,
            mode='rb',
        ) as file_object:
            contents = await file_object.read()
            return await self.upload(bucket, object_name, contents,
                                     **kwargs)

    @staticmethod
    def _get_stream_len(stream: IO[AnyStr]) -> int:
        current = stream.tell()
        try:
            return stream.seek(0, os.SEEK_END)
        finally:
            stream.seek(current)

    @staticmethod
    def _preprocess_data(data: Any) -> IO[Any]:
        if data is None:
            return io.StringIO('')

        if isinstance(data, bytes):
            return io.BytesIO(data)
        if isinstance(data, str):
            return io.StringIO(data)
        if isinstance(data, io.IOBase):
            return data  # type: ignore[return-value]

        raise TypeError(f'unsupported upload type: "{type(data)}"')

    @staticmethod
    def _decide_upload_type(force_resumable_upload: Optional[bool],
                            content_length: int) -> UploadType:
        # force resumable
        if force_resumable_upload is True:
            return UploadType.RESUMABLE

        # force simple
        if force_resumable_upload is False:
            return UploadType.SIMPLE

        # decide based on Content-Length
        if content_length > MAX_CONTENT_LENGTH_SIMPLE_UPLOAD:
            return UploadType.RESUMABLE

        return UploadType.SIMPLE

    @staticmethod
    def _split_content_type(content_type: str) -> Tuple[str, Optional[str]]:
        content_type_and_encoding_split = content_type.split(';')
        content_type = content_type_and_encoding_split[0].lower().strip()

        encoding = None
        if len(content_type_and_encoding_split) > 1:
            encoding_str = content_type_and_encoding_split[1].lower().strip()
            encoding = encoding_str.split('=')[-1]

        return content_type, encoding

    @staticmethod
    def _format_metadata_key(key: str) -> str:
        """
        Formats the fixed-key metadata keys as wanted by the multipart API.

        Ex: Content-Disposition --> contentDisposition
        """
        parts = key.split('-')
        parts = [parts[0].lower()] + [p.capitalize() for p in parts[1:]]
        return ''.join(parts)

    async def _download(self, bucket: str, object_name: str, *,
                        params: Optional[Dict[str, str]] = None,
                        headers: Optional[Dict[str, str]] = None,
                        timeout: int = DEFAULT_TIMEOUT,
                        session: Optional[Session] = None) -> bytes:
        # https://cloud.google.com/storage/docs/request-endpoints#encoding
        encoded_object_name = quote(object_name, safe='')
        url = f'{API_ROOT}/{bucket}/o/{encoded_object_name}'
        headers = headers or {}
        headers.update(await self._headers())

        s = AioSession(session) if session else self.session
        response = await s.get(url, headers=headers, params=params or {},
                               timeout=timeout)

        # N.B. the GCS API sometimes returns 'application/octet-stream' when a
        # string was uploaded. To avoid potential weirdness, always return a
        # bytes object.
        try:
            data: bytes = await response.read()
        except (AttributeError, TypeError):
            data = response.content

        return data

    async def _download_stream(self, bucket: str, object_name: str, *,
                               params: Optional[Dict[str, str]] = None,
                               headers: Optional[Dict[str, str]] = None,
                               timeout: int = DEFAULT_TIMEOUT,
                               session: Optional[Session] = None
                               ) -> StreamResponse:
        # https://cloud.google.com/storage/docs/request-endpoints#encoding
        encoded_object_name = quote(object_name, safe='')
        url = f'{API_ROOT}/{bucket}/o/{encoded_object_name}'
        headers = headers or {}
        headers.update(await self._headers())

        s = AioSession(session) if session else self.session

        if BUILD_GCLOUD_REST:
            # stream argument is only expected by requests.Session.
            # pylint: disable=unexpected-keyword-arg
            return StreamResponse(s.get(url, headers=headers,
                                        params=params or {},
                                        timeout=timeout, stream=True))
        return StreamResponse(await s.get(url, headers=headers,
                                          params=params or {},
                                          timeout=timeout))

    async def _upload_simple(self, url: str, object_name: str,
                             stream: IO[AnyStr], params: Dict[str, str],
                             headers: Dict[str, str], *,
                             session: Optional[Session] = None,
                             timeout: int = 30) -> Dict[str, Any]:
        # https://cloud.google.com/storage/docs/json_api/v1/how-tos/simple-upload
        params['name'] = object_name
        params['uploadType'] = 'media'

        s = AioSession(session) if session else self.session
        resp = await s.post(url, data=stream, headers=headers, params=params,
                            timeout=timeout)
        data: Dict[str, Any] = await resp.json(content_type=None)
        return data

    async def _upload_multipart(self, url: str, object_name: str,
                                stream: IO[AnyStr], params: Dict[str, str],
                                headers: Dict[str, str],
                                metadata: Dict[str, Any], *,
                                session: Optional[Session] = None,
                                timeout: int = 30) -> Dict[str, Any]:
        # https://cloud.google.com/storage/docs/json_api/v1/how-tos/multipart-upload
        params['uploadType'] = 'multipart'

        metadata_headers = {'Content-Type': 'application/json; charset=UTF-8'}
        metadata = {self._format_metadata_key(k): v
                    for k, v in metadata.items()}
        if 'metadata' in metadata:
            metadata['metadata'] = {
                str(k): str(v) if v is not None else None
                for k, v in metadata['metadata'].items()}

        metadata['name'] = object_name

        raw_body: AnyStr = stream.read()
        if isinstance(raw_body, str):
            bytes_body: bytes = raw_body.encode('utf-8')
        else:
            bytes_body = raw_body

        parts = [
            (metadata_headers, json.dumps(metadata).encode('utf-8')),
            ({'Content-Type': headers['Content-Type']}, bytes_body),
        ]
        boundary = choose_boundary()
        body, content_type = encode_multipart_formdata(parts, boundary)
        headers.update({
            'Content-Type': content_type,
            'Content-Length': str(len(body)),
            'Accept': 'application/json'
        })

        s = AioSession(session) if session else self.session
        resp = await s.post(url, data=body, headers=headers, params=params,
                            timeout=timeout)
        data: Dict[str, Any] = await resp.json(content_type=None)
        return data

    async def _upload_resumable(self, url: str, object_name: str,
                                stream: IO[AnyStr], params: Dict[str, str],
                                headers: Dict[str, str], *,
                                metadata: Optional[Dict[str, Any]] = None,
                                session: Optional[Session] = None,
                                timeout: int = 30) -> Dict[str, Any]:
        # https://cloud.google.com/storage/docs/json_api/v1/how-tos/resumable-upload
        session_uri = await self._initiate_upload(url, object_name, params,
                                                  headers, metadata=metadata,
                                                  session=session)
        return await self._do_upload(session_uri, stream, headers=headers,
                                     session=session, timeout=timeout)

    async def _initiate_upload(self, url: str, object_name: str,
                               params: Dict[str, str], headers: Dict[str, str],
                               *, metadata: Optional[Dict[str, Any]] = None,
                               timeout: int = DEFAULT_TIMEOUT,
                               session: Optional[Session] = None) -> str:
        params['uploadType'] = 'resumable'

        metadict = (metadata or {}).copy()
        metadict = {self._format_metadata_key(k): v
                    for k, v in metadict.items()}
        if 'metadata' in metadict:
            metadict['metadata'] = {
                str(k): str(v) if v is not None else None
                for k, v in metadict['metadata'].items()}

        metadict.update({'name': object_name})
        metadata_ = json.dumps(metadict)

        post_headers = headers.copy()
        post_headers.update({
            'Content-Length': str(len(metadata_)),
            'Content-Type': 'application/json; charset=UTF-8',
            'X-Upload-Content-Type': headers['Content-Type'],
            'X-Upload-Content-Length': headers['Content-Length']
        })

        s = AioSession(session) if session else self.session
        resp = await s.post(url, headers=post_headers, params=params,
                            data=metadata_, timeout=timeout)
        session_uri: str = resp.headers['Location']
        return session_uri

    async def _do_upload(self, session_uri: str, stream: IO[AnyStr],
                         headers: Dict[str, str], *, retries: int = 5,
                         session: Optional[Session] = None,
                         timeout: int = 30) -> Dict[str, Any]:
        s = AioSession(session) if session else self.session

        for tries in range(retries):
            try:
                resp = await s.put(session_uri, headers=headers,
                                   data=stream, timeout=timeout)
            except ResponseError:
                headers.update({'Content-Range': '*/*'})
                await sleep(2. ** tries)  # type: ignore[func-returns-value]

                continue

            break

        data: Dict[str, Any] = await resp.json(content_type=None)
        return data

    async def patch_metadata(
            self, bucket: str, object_name: str, metadata: Dict[str, Any],
            *, params: Optional[Dict[str, str]] = None,
            headers: Optional[Dict[str, str]] = None,
            session: Optional[Session] = None,
            timeout: int = DEFAULT_TIMEOUT) -> Dict[str, Any]:
        # https://cloud.google.com/storage/docs/json_api/v1/objects/patch
        encoded_object_name = quote(object_name, safe='')
        url = f'{API_ROOT}/{bucket}/o/{encoded_object_name}'
        params = params or {}
        headers = headers or {}
        headers.update(await self._headers())
        headers['Content-Type'] = 'application/json'

        s = AioSession(session) if session else self.session
        resp = await s.patch(url, data=json.dumps(metadata).encode('utf-8'),
                             headers=headers, params=params, timeout=timeout)
        data: Dict[str, Any] = await resp.json(content_type=None)
        return data

    async def get_bucket_metadata(self, bucket: str, *,
                                  params: Optional[Dict[str, str]] = None,
                                  headers: Optional[Dict[str, str]] = None,
                                  session: Optional[Session] = None,
                                  timeout: int = DEFAULT_TIMEOUT
                                  ) -> Dict[str, Any]:
        url = f'{API_ROOT}/{bucket}'
        headers = headers or {}
        headers.update(await self._headers())

        s = AioSession(session) if session else self.session
        resp = await s.get(url, headers=headers, params=params or {},
                           timeout=timeout)
        data: Dict[str, Any] = await resp.json(content_type=None)
        return data

    async def close(self) -> None:
        await self.session.close()

    async def __aenter__(self) -> 'Storage':
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.close()



from pkg_resources import get_distribution
__version__ = get_distribution('gcloud-aio-storage').version

from gcloud.aio.storage.blob import Blob
from gcloud.aio.storage.bucket import Bucket
from gcloud.aio.storage.storage import SCOPES
from gcloud.aio.storage.storage import Storage
from gcloud.aio.storage.storage import StreamResponse


__all__ = ['__version__', 'Blob', 'Bucket', 'SCOPES', 'Storage',
           'StreamResponse']



import io
import os
import random
import string
import uuid

import pytest
from gcloud.aio.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.aio.storage import Storage

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session

# TODO: use hypothesis
RANDOM_BINARY = os.urandom(2045)

# Updated statement to make it compatible with python2
rand_str_list = [random.choice(string.ascii_letters) for r in range(0, 1054)]
RANDOM_STRING = ''.join(rand_str_list)


@pytest.mark.asyncio
@pytest.mark.parametrize('uploaded_data,expected_data', [
    (io.BytesIO(RANDOM_BINARY), RANDOM_BINARY),
    (io.StringIO(RANDOM_STRING), RANDOM_STRING.encode('utf-8')),
])
async def test_download_stream(bucket_name, creds, uploaded_data,
                               expected_data):
    object_name = f'{uuid.uuid4().hex}/{uuid.uuid4().hex}'

    async with Session() as session:
        storage = Storage(service_file=creds, session=session)
        res = await storage.upload(bucket_name, object_name, uploaded_data)

        with io.BytesIO(b'') as downloaded_data:
            download_stream = await storage.download_stream(
                bucket_name, res['name'])
            while True:
                chunk = await download_stream.read(4096)
                if not chunk:
                    break
                downloaded_data.write(chunk)

            assert expected_data == downloaded_data.getvalue()

        await storage.delete(bucket_name, res['name'])



import io
import json
import os
import random
import string
import uuid

import pytest
from gcloud.aio.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.aio.storage import Storage

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session

# TODO: use hypothesis
RANDOM_BINARY = os.urandom(2045)

# Updated statement to make it compatible with python2
rand_str_list = [random.choice(string.ascii_letters) for r in range(0, 1054)]
RANDOM_STRING = ''.join(rand_str_list)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'uploaded_data,range_header,expected_data,file_extension', [
        (json.dumps([1, 2, 3]), 'bytes=0-1', json.dumps(
            [1, 2, 3]).encode('utf-8')[0:2], 'json'),
        ('test'.encode('utf-8'), 'bytes=2-3', 'test'.encode('utf-8')[2:4],
         'bin'),
        (io.BytesIO(RANDOM_BINARY), 'bytes=1-1000', RANDOM_BINARY[1:1001],
         'bin'),
        (io.StringIO(RANDOM_STRING), 'bytes=10-100',
         RANDOM_STRING.encode('utf-8')[10:101], 'txt'),
    ])
async def test_download_range(bucket_name, creds, uploaded_data, range_header,
                              expected_data, file_extension):
    object_name = f'{uuid.uuid4().hex}/{uuid.uuid4().hex}.{file_extension}'

    async with Session() as session:
        storage = Storage(service_file=creds, session=session)
        res = await storage.upload(bucket_name, object_name, uploaded_data)

        downloaded_data = await storage.download(
            bucket_name, res['name'], headers={'Range': range_header})
        assert expected_data == downloaded_data

        await storage.delete(bucket_name, res['name'])



import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import TYPE_CHECKING

from gcloud.aio.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module

from .blob import Blob

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import HTTPError as ResponseError
    from requests import Session
else:
    from aiohttp import ClientResponseError as ResponseError  # type: ignore[no-redef]  # pylint: disable=line-too-long
    from aiohttp import ClientSession as Session  # type: ignore[no-redef]

if TYPE_CHECKING:
    from .storage import Storage  # pylint: disable=cyclic-import


log = logging.getLogger(__name__)


class Bucket:
    def __init__(self, storage: 'Storage', name: str) -> None:
        self.storage = storage
        self.name = name

    async def get_blob(self, blob_name: str,
                       session: Optional[Session] = None) -> Blob:
        metadata = await self.storage.download_metadata(self.name, blob_name,
                                                        session=session)

        return Blob(self, blob_name, metadata)

    async def blob_exists(self, blob_name: str,
                          session: Optional[Session] = None) -> bool:
        try:
            await self.get_blob(blob_name, session=session)
            return True
        except ResponseError as e:
            try:
                if e.status in {404, 410}:  # type: ignore[attr-defined]
                    return False
            except AttributeError:
                if e.code in {404, 410}:  # type: ignore[attr-defined]
                    return False

            raise e

    async def list_blobs(self, prefix: str = '',
                         session: Optional[Session] = None) -> List[str]:
        params = {'prefix': prefix, 'pageToken': ''}
        items = []
        while True:
            content = await self.storage.list_objects(self.name,
                                                      params=params,
                                                      session=session)
            items.extend([x['name'] for x in content.get('items', list())])

            params['pageToken'] = content.get('nextPageToken', '')
            if not params['pageToken']:
                break

        return items

    def new_blob(self, blob_name: str) -> Blob:
        return Blob(self, blob_name, {'size': 0})

    async def get_metadata(
            self, params: Optional[Dict[str, Any]] = None,
            session: Optional[Session] = None
    ) -> Dict[str, Any]:
        return await self.storage.get_bucket_metadata(self.name, params=params,
                                                      session=session)



import binascii
import collections
import datetime
import enum
import hashlib
import io
from typing import Any
from typing import Dict
from typing import Optional
from typing import TYPE_CHECKING
from urllib.parse import quote

import rsa
from gcloud.aio.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.aio.auth import Token  # pylint: disable=no-name-in-module
from gcloud.aio.storage.constants import DEFAULT_TIMEOUT
from pyasn1.codec.der import decoder
from pyasn1_modules import pem
from pyasn1_modules.rfc5208 import PrivateKeyInfo

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session  # type: ignore[no-redef]

if TYPE_CHECKING:
    from .bucket import Bucket  # pylint: disable=cyclic-import


HOST = 'storage.googleapis.com'

PKCS1_MARKER = ('-----BEGIN RSA PRIVATE KEY-----',
                '-----END RSA PRIVATE KEY-----')
PKCS8_MARKER = ('-----BEGIN PRIVATE KEY-----',
                '-----END PRIVATE KEY-----')
PKCS8_SPEC = PrivateKeyInfo()


class PemKind(enum.Enum):
    """
    Tracks the response of ``pem.readPemBlocksFromFile(key, *args)``>

    Note that the specified method returns ``(marker_id, key_bytes)``, where
    ``marker_id`` is the integer index of the matching ``arg`` (or -1 if no
    match was found.

    For example::

        (marker_id, _) = pem.readPemBlocksFromFile(key, PKCS1_MARKER,
                                                   PCKS8_MARKER)
        if marker_id == -1:
            # "key" did not match either type or was invalid
        if marker_id == 0:
            # "key" matched the zeroth provided marker arg, eg. PKCS1_MARKER
        if marker_id == 1:
            # "key" matched the zeroth provided marker arg, eg. PKCS8_MARKER
    """
    INVALID = -1
    PKCS1 = 0
    PKCS8 = 1


class Blob:
    def __init__(self, bucket: 'Bucket', name: str,
                 metadata: Dict[str, Any]) -> None:
        self.__dict__.update(**metadata)

        self.bucket = bucket
        self.name = name
        self.size: int = int(self.size)

    @property
    def chunk_size(self) -> int:
        return self.size + (262144 - (self.size % 262144))

    async def download(self, timeout: int = DEFAULT_TIMEOUT,
                       session: Optional[Session] = None) -> Any:
        return await self.bucket.storage.download(self.bucket.name,
                                                  self.name,
                                                  timeout=timeout,
                                                  session=session)

    async def upload(self, data: Any,
                     session: Optional[Session] = None) -> Dict[str, Any]:
        metadata = await self.bucket.storage.upload(
            self.bucket.name, self.name, data, session=session)

        self.__dict__.update(metadata)
        return metadata

    async def get_signed_url(  # pylint: disable=too-many-locals
            self, expiration: int, headers: Optional[Dict[str, str]] = None,
            query_params: Optional[Dict[str, Any]] = None,
            http_method: str = 'GET', token: Optional[Token] = None) -> str:
        """
        Create a temporary access URL for Storage Blob accessible by anyone
        with the link.

        Adapted from Google Documentation:
        https://cloud.google.com/storage/docs/access-control/signing-urls-manually#python-sample
        """
        if expiration > 604800:
            raise ValueError("expiration time can't be longer than 604800 "
                             'seconds (7 days)')

        quoted_name = quote(self.name, safe=b'/~')
        canonical_uri = f'/{quoted_name}'

        datetime_now = datetime.datetime.utcnow()
        request_timestamp = datetime_now.strftime('%Y%m%dT%H%M%SZ')
        datestamp = datetime_now.strftime('%Y%m%d')

        # TODO: support for GCE_METADATA and AUTHORIZED_USER tokens. It should
        # be possible via API, but seems to not be working for us in some
        # cases.
        #
        # https://cloud.google.com/iam/docs/reference/credentials/rest/v1/
        # projects.serviceAccounts/signBlob
        token = token or self.bucket.storage.token
        client_email = token.service_data.get('client_email')
        private_key = token.service_data.get('private_key')
        if not client_email or not private_key:
            raise KeyError('Blob signing is only suported for tokens with '
                           'explicit client_email and private_key data; '
                           'please check your token points to a JSON service '
                           'account file')

        credential_scope = f'{datestamp}/auto/storage/goog4_request'
        credential = f'{client_email}/{credential_scope}'

        headers = headers or {}
        headers['host'] = f'{self.bucket.name}.{HOST}'

        ordered_headers = collections.OrderedDict(sorted(headers.items()))
        canonical_headers = ''.join(
            f'{str(k).lower()}:{str(v).lower()}\n'
            for k, v in ordered_headers.items())

        signed_headers = ';'.join(
            f'{str(k).lower()}' for k in ordered_headers.keys())

        query_params = query_params or {}
        query_params['X-Goog-Algorithm'] = 'GOOG4-RSA-SHA256'
        query_params['X-Goog-Credential'] = credential
        query_params['X-Goog-Date'] = request_timestamp
        query_params['X-Goog-Expires'] = expiration
        query_params['X-Goog-SignedHeaders'] = signed_headers

        ordered_query_params = collections.OrderedDict(
            sorted(query_params.items()))

        canonical_query_str = '&'.join(
            f'{quote(str(k), safe="")}={quote(str(v), safe="")}'
            for k, v in ordered_query_params.items())

        canonical_req = '\n'.join([http_method, canonical_uri,
                                   canonical_query_str, canonical_headers,
                                   signed_headers, 'UNSIGNED-PAYLOAD'])
        canonical_req_hash = hashlib.sha256(canonical_req.encode()).hexdigest()

        str_to_sign = '\n'.join(['GOOG4-RSA-SHA256', request_timestamp,
                                 credential_scope, canonical_req_hash])

        # N.B. see the ``PemKind`` enum
        marker_id, key_bytes = pem.readPemBlocksFromFile(
            io.StringIO(private_key), PKCS1_MARKER, PKCS8_MARKER)
        if marker_id == PemKind.INVALID.value:
            raise ValueError('private key is invalid or unsupported')

        if marker_id == PemKind.PKCS8.value:
            # convert from pkcs8 to pkcs1
            key_info, remaining = decoder.decode(key_bytes,
                                                 asn1Spec=PKCS8_SPEC)
            if remaining != b'':
                raise ValueError('could not read PKCS8 key: found extra bytes',
                                 remaining)

            private_key_info = key_info.getComponentByName('privateKey')
            key_bytes = private_key_info.asOctets()

        key = rsa.key.PrivateKey.load_pkcs1(key_bytes, format='DER')
        signed_blob = rsa.pkcs1.sign(str_to_sign.encode(), key, 'SHA-256')

        signature = binascii.hexlify(signed_blob).decode()

        return (f'https://{self.bucket.name}.{HOST}{canonical_uri}?'
                f'{canonical_query_str}&X-Goog-Signature={signature}')



import os

import pytest


@pytest.fixture(scope='module')
def bucket_name() -> str:
    return 'dialpad-oss-public-test'


@pytest.fixture(scope='module')
def creds() -> str:
    # TODO: bundle public creds into this repo
    return os.environ['GOOGLE_APPLICATION_CREDENTIALS']



import io
import json
import os
import random
import string
import uuid

import pytest
from gcloud.aio.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.aio.storage import Storage

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session

# TODO: use hypothesis
RANDOM_BINARY = os.urandom(2045)

# Updated statement to make it compatible with python2
rand_str_list = [random.choice(string.ascii_letters) for r in range(0, 1054)]
RANDOM_STRING = ''.join(rand_str_list)


@pytest.mark.asyncio
@pytest.mark.parametrize('uploaded_data,expected_data,file_extension', [
    ('test', b'test', 'txt'),
    (json.dumps({'data': 1}), json.dumps({'data': 1}).encode('utf-8'), 'json'),
    (json.dumps([1, 2, 3]), json.dumps([1, 2, 3]).encode('utf-8'), 'json'),
    ('test'.encode('utf-8'), 'test'.encode('utf-8'), 'bin'),
    (io.BytesIO(RANDOM_BINARY), RANDOM_BINARY, 'bin'),
    (io.StringIO(RANDOM_STRING), RANDOM_STRING.encode('utf-8'), 'txt'),
])
async def test_upload_multipart(bucket_name, creds, uploaded_data,
                                expected_data, file_extension):
    object_name = f'{uuid.uuid4().hex}/{uuid.uuid4().hex}.{file_extension}'

    async with Session() as session:
        storage = Storage(service_file=creds, session=session)
        res = await storage.upload(
            bucket_name,
            object_name,
            uploaded_data,
            metadata={'Content-Disposition': 'inline',
                      'metadata': {'a': 1, 'b': 2}})

        try:
            assert res['name'] == object_name

            downloaded_data = await storage.download(bucket_name, res['name'])
            assert downloaded_data == expected_data

            downloaded_metadata = await storage.download_metadata(bucket_name,
                                                                  res['name'])
            assert downloaded_metadata.pop('contentDisposition') == 'inline'
            assert downloaded_metadata['metadata']['a'] == '1'
            assert downloaded_metadata['metadata']['b'] == '2'

        finally:
            # TODO: don't bother
            # await storage.delete(bucket_name, res['name'])
            pass
